/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.http;

import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.apache.axiom.om.OMOutputFormat;
import org.apache.axiom.soap.SOAPEnvelope;
import org.apache.axiom.soap.SOAPFault;
import org.apache.axiom.soap.SOAPFaultCode;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.Constants;
import org.bluestemsoftware.specification.eoa.component.binding.rt.BindingFaultRT;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAP11HTTPBindingRT;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPBindingConfiguration;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPSenderFault;
import org.bluestemsoftware.specification.eoa.ext.feature.http.server.HTTPServerRequest;

public final class SOAP11HTTPBindingImpl extends SOAPHTTPBindingImpl implements SOAP11HTTPBindingRT.Provider {

    public SOAP11HTTPBindingImpl(SOAPBindingConfiguration configuration, Set<String> features) {
        super(configuration, features);
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.http.SOAPHTTPProtocolImpl#getContentType(org.apache.axiom.om.OMOutputFormat,
     *      java.lang.String)
     */
    @Override
    public String getContentType(OMOutputFormat omFormat, String soapAction) {
        // note that soap action is manifested as an http header if soap 1.1
        String contentType = omFormat.getContentType();
        String encoding = omFormat.getCharSetEncoding();
        if (encoding != null) {
            contentType = contentType + "; charset=" + encoding;
        }
        return contentType;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.http.SOAPHTTPProtocolImpl#getSOAPAction(javax.servlet.http.HTTPServerRequest)
     */
    protected String getSOAPAction(HTTPServerRequest request) {
        String soapAction = request.getHeader(Constants.SOAP_ACTION_HEADER);
        if (soapAction == null || soapAction.equals("")) {
            return null;
        } else {
            soapAction = soapAction.trim();
        }
        if (soapAction.startsWith("\"") && soapAction.endsWith("\"")) {
            // remove surrounding quotes and trim any spaces inside quoted string
            soapAction = soapAction.substring(1, soapAction.length() - 1).trim();
        }
        return soapAction;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.http.SOAPHTTPProtocolImpl#getMediaType(HTTPServerRequest)
     */
    @Override
    protected String getMediaType(String contentType) throws BindingFaultRT {
        String mediaType = null;
        int index = contentType.indexOf(';');
        if (index > 0) {
            mediaType = contentType.substring(0, index);
        } else {
            mediaType = contentType;
        }
        if (mediaType.equals(Constants.SOAP_11_MEDIA_TYPE) || mediaType.equals(Constants.SOAP_MTOM_MEDIA_TYPE)) {
            return mediaType;
        } else {
            throw new SOAPSenderFault(getSOAPVersion(), "Found media type '"
                    + mediaType
                    + "'. Expected "
                    + Constants.SOAP_11_MEDIA_TYPE);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.http.SOAPHTTPBindingImpl#setSOAPAction(java.util.Map)
     */
    @Override
    protected String setSOAPAction(Map<String, String> httpHeaders) {

        // soap 1.1 binding rules indicates that a soap action must be
        // supplied and that empty string is valid. ws-a indicates that
        // if a non-empty value is specified it must be the wsa:Action 
        // value. so ... we will play it safe and always pass the empty
        // string regardless of what was defined via wsdl

        httpHeaders.put(Constants.SOAP_ACTION_HEADER, "");

        return "";

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.http.SOAPHTTPBindingImpl#isSOAPProtocolFault(org.apache.axiom.soap.SOAPEnvelope)
     */
    @Override
    protected boolean isSOAPProtocolFault(SOAPEnvelope soapEnvelope) {
        if (soapEnvelope == null) {
            return false;
        }
        if (soapEnvelope.getBody().hasFault()) {
            SOAPFault sf = soapEnvelope.getBody().getFault();
            SOAPFaultCode code = sf.getCode();
            QName name = code.getTextAsQName();
            if (name.equals(Constants.SOAP_11_MUST_UNDERSTAND_FAULT)) {
                return true;
            } else if (name.equals(Constants.SOAP_11_VERSION_MISMATCH_FAULT)) {
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

}
