/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util;

import java.io.IOException;
import java.io.InputStream;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointActionReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointActionReference.RequestActionReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointActionReference.ResponseActionReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceFaultReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceMessageReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceMessageReference.ContentModel;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext.MessageModules;
import org.bluestemsoftware.specification.eoa.component.message.InterfaceMessage;
import org.bluestemsoftware.specification.eoa.component.message.MessagePart;
import org.bluestemsoftware.specification.eoa.component.message.rt.Content;
import org.bluestemsoftware.specification.eoa.component.message.rt.Message;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPDocumentStyleMessage;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPHeaderBlock;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessage;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessageStyle;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPVersion;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPFault;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPReceiverFault;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPSenderFault;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.appdata.ApplicationData;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.appdata.ApplicationDataModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.WSAMessageModule;
import org.bluestemsoftware.specification.eoa.system.SystemContext;
import org.bluestemsoftware.specification.eoa.system.System.Log;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

public class SOAPReader {

    private static final Log log = SystemContext.getContext().getSystem().getLog(SOAPReader.class);

    public static Message readMessage(SOAPVersion sv, EndpointReference epr, InterfaceMessage am, SOAPMessage cm, MessageModules messageModules, InputStream in) throws SOAPFault {
        log.trace("readMessage begin");
        
        WSAMessageModule wsamm = messageModules.getModule(WSAMessageModule.class);
        EndpointActionReference ear = epr.getEndpointActionReference(wsamm.getAction());

        Message message = null;

        if (ear != null) {

            // the action is declared, i.e. it's an action with direction 'in' for
            // my service, or it's a declared action with direction 'out' for a
            // partner service, i.e. a normal partner response message or a declared
            // fault response

            if (ear instanceof RequestActionReference) {
                message = ((RequestActionReference)ear).createAction().getMessage();
            } else {
                String rt = wsamm.getRelatesTo();
                message = ((ResponseActionReference)ear).createAction(rt).getMessage();
            }

        } else {

            // it's an undeclared partner response action which was converted to
            // a system fault message by binding, i.e. content is a runtime
            // instance of system fault message. clone definition and overwrite
            // payload with parsed content

            message = ((Message)am.getRuntimeProvider()).cloneDefinition();

            Content content = null;
            try {
                content = message.parseContent(in);
            } catch (Exception ex) {
                log.error("Error reading partner response. " + ex);
                throw new SOAPReceiverFault(sv, "Error reading partner response.");
            }

            message.setContent(content);

            return message;

        }

        InterfaceAction ia = ear.getReferencedComponent().getInterfaceAction();

        Content content = null;
        if (message != null) {
            try {
                content = message.parseContent(in);
            } catch (Throwable th) {
                log.error("Error reading soap:Body content. " + th);
                throw new SOAPReceiverFault(sv, "Error reading soap:Body content.");
            }
        } else {
            try {
                if (in.read() > -1) {
                    throw new SOAPSenderFault(sv, "Expected an empty payload, but payload has content.");
                }
            } catch (IOException ignore) {
            }
        }

        // note that according to basic profile (R2724) 'in all other cases
        // where an envelope is inconsistent with its WSDL description, a
        // fault with a 'Client' faultcode should be generated'. so, we'll
        // throw soap sender faults from this point forward

        boolean implementDefaultBindingRules = false;

        // if concrete message is null, either the binding is re-usable or
        // the operation is unbound, i.e. wsdl 1.1 requires a concrete
        // message def. this implies that we implement default wsdl 2.0
        // binding rules

        if (cm == null) {
            implementDefaultBindingRules = true;
        }

        // if action is a fault action, wsdl 1.1 requires that message have
        // a single part typed as an element and no headers. wsdl 2.0 does
        // not allow user to define content model for faults, i.e. #element
        // is implied and although headers can be declared, wsdl 2.0 never
        // maps them to a message part. so default wsdl 2.0 binding rules
        // work in either case

        if (ia instanceof InterfaceFaultReference) {
            implementDefaultBindingRules = true;
        }

        // so according to the rules defined above, if indicated, we implement
        // default binding rules and return

        if (implementDefaultBindingRules) {
            return implementDefaultBindingRules(sv, ia, am, message, content);
        }

        // if the abstract message is null then it's a wsdl 2.0 style normal
        // message, i.e. wsdl 1.1 requires an abstract message definition.
        // the content model must be either #none or #any. an exception was
        // thrown during validation if #other was specified

        if (am == null) {

            ContentModel contentModel = ((InterfaceMessageReference)ia).getContentModel();
            if (contentModel == ContentModel.ANY) {
                if (content == null) {
                    throw new SOAPSenderFault(sv, "ContentModel #any, but payload empty.");
                } else {
                    message.setContent(content);
                }
            } else {
                if (content != null) {
                    throw new SOAPSenderFault(sv, "Expected an empty payload, but payload has content.", content);
                }
            }

            return message;

        }

        Message result = null;
        if (cm.getMessageStyle() == SOAPMessageStyle.DOCUMENT_STYLE) {
            result = readDocumentStyleMessage(sv, ia, cm, am, message, messageModules, content);
        } else {
            result = readWSDL11RPCStyleMessage(sv, ia, cm, am, message, messageModules, content);
        }

        log.trace("readMessage end");
        return result;
    }

    /*
     * Implements WSDL 2.0 default binding rules. Note that logic is only executed if binding
     * definition was imported via WSDL 2.0, i.e. when concrete message definition is null. A
     * concrete message def is absolutely required for bindings described via WSDL 1.1.
     */
    private static Message implementDefaultBindingRules(SOAPVersion sv, InterfaceAction ia, InterfaceMessage am, Message message, Content content) throws SOAPFault {
        log.trace("implementDefaultBindingRules begin");
        
        if (ia instanceof InterfaceMessageReference) {

            // it's a normal message. wsdl 20 allows for different content
            // models which we must enforce. note that an exception was
            // thrown during validation if #other is specified

            ContentModel contentModel = ((InterfaceMessageReference)ia).getContentModel();
            if (contentModel == ContentModel.NONE) {
                if (content != null) {
                    throw new SOAPSenderFault(sv, "Expected an empty payload, but payload has content.", content);
                }
            } else if (contentModel == ContentModel.ANY) {
                if (content == null) {
                    throw new SOAPSenderFault(sv, "ContentModel #any, but payload empty.");
                } else {
                    message.setContent(content);
                }
            } else {
                if (content == null) {
                    throw new SOAPSenderFault(sv, "Invalid payload. Expected type "
                            + am.getName()
                            + ", but payload empty.");
                }
                QName payloadName = new QName(content.getNamespaceURI(), content.getLocalName());
                if (!payloadName.equals(am.getName())) {
                    throw new SOAPSenderFault(sv, "Invalid payload. Expected type "
                            + am.getName()
                            + ", but found type "
                            + payloadName
                            + ".", content);
                } else {
                    message.setContent(content);
                }
            }

        } else {

            // it's a fault message. content model is undefined, but is assumed to
            // be #element. note that content represents content extracted from
            // soap version specific fault detail element by binding impl

            if (content == null) {
                throw new SOAPSenderFault(sv, "Invalid fault message. Expected type "
                        + am.getName()
                        + ", but payload empty.");
            }

            QName payloadName = new QName(content.getNamespaceURI(), content.getLocalName());
            if (!payloadName.equals(am.getName())) {
                throw new SOAPSenderFault(sv, "Invalid fault message. Expected type "
                        + am.getName()
                        + ", but found type "
                        + payloadName
                        + ".", content);
            } else {
                message.setContent(content);
            }

        }

        // we're done implementing default binding rules. note that
        // there are no declared headers, i.e. no concrete message
        // definition -> no declared headers

        log.trace("implementDefaultBindingRules end");
        return message;

    }

    /*
     * Processes a 'normal' document style message, i.e. fault messages are processed within
     * implementDefaultBindingRules method.
     */
    private static Message readDocumentStyleMessage(SOAPVersion sv, InterfaceAction ia, SOAPMessage cm, InterfaceMessage am, Message message, MessageModules messageModules, Content content) throws SOAPFault {
        log.trace("readDocumentStyleMessage begin");
        
        // if content model is defined, it's a wsdl 2.0 style message, i.e.
        // a single abstract part typed as an element which is msg payload.
        // if content model is null, it's a wsdl 1.1 document style msg.
        // note that an exception was thrown during validation if #other is
        // specified

        Message result = null;
        
        ContentModel contentModel = ((InterfaceMessageReference)ia).getContentModel();
        if (contentModel != null) {
            if (contentModel != ContentModel.ELEMENT) {
                throw new SOAPSenderFault(sv, "Expected ContentModel #element. Found " + contentModel);
            }
            QName payloadName = new QName(content.getNamespaceURI(), content.getLocalName());
            if (!payloadName.equals(am.getName())) {
                throw new SOAPSenderFault(sv, "Invalid payload. Expected type "
                        + am.getName()
                        + ", but found type "
                        + payloadName
                        + ".", content);
            } else {
                message.setContent(content);
                result = message;
            }
        } else {
            result = readWSDL11DocumentStyleMessage(sv, ia, cm, am, message, messageModules, content);
        }

        log.trace("readDocumentStyleMessage end");
        return result;
    }

    /*
     * Reads a WSDL 1.1 document style message, i.e. message payload is an element. Note that
     * unlike a WSDL 2.0 style message, a WSDL 1.1 document style message MAY have multiple
     * parts, i.e. parts that are mapped soap header block elements. Note that only normal
     * messages are processed by this method, i.e. fault messages are processed within
     * implementDefaultBindingRules method.
     */
    private static Message readWSDL11DocumentStyleMessage(SOAPVersion sv, InterfaceAction ia, SOAPMessage cm, InterfaceMessage am, Message message, MessageModules messageModules, Content content) throws SOAPFault {
        log.trace("readWSDL11DocumentStyleMessage begin");
        
        // wsdl 2.0 uses content model to declare an empty payload. content model
        // is undefined for wsdl 1.1, however, so check for a null soap body part

        String soapBodyPart = ((SOAPDocumentStyleMessage)cm).getBodyPart();

        if (soapBodyPart != null) {

            // retrieve abstract message part mapped to soap body and verify that its
            // type is as expected

            MessagePart abstractPart = am.getPart(soapBodyPart);
            QName elementType = abstractPart.getSchemaComponentQName();

            if (content == null) {
                throw new SOAPSenderFault(sv, "Invalid payload. Expected type "
                        + elementType
                        + ", but payload empty.");
            }

            QName payloadType = new QName(content.getNamespaceURI(), content.getLocalName());

            if (!payloadType.equals(elementType)) {
                throw new SOAPSenderFault(sv, "Invalid payload. Expected type "
                        + elementType
                        + ", but found type "
                        + payloadType
                        + ".", content);
            } else {
                message.setContent(soapBodyPart, content);
            }

        } else {

            if (content != null) {
                throw new SOAPSenderFault(sv, "Expected an empty payload, but payload has content.");
            }

        }

        // now we process declared soap headers, if any, each of which maps to a
        // a part declared on abstract message. application data module contains
        // headers extracted from message targeted to ultimate destination, not
        // consumed by binding modules

        ApplicationDataModule adm = messageModules.getModule(ApplicationDataModule.class);
        Map<QName, Set<ApplicationData>> ad = adm.getApplicationData();

        for (SOAPHeaderBlock hb : cm.getHeaders()) {

            // retrieve header block type. note that headers are always typed
            // as an element. basic profile validated abstract and concrete
            // message parts are in sync

            String soapHeaderPart = hb.getPartName();
            MessagePart abstractPart = am.getPart(soapHeaderPart);
            QName elementType = abstractPart.getSchemaComponentQName();
            
            // retrieve application data. note that we know a header block
            // of required type exists because required headers check has
            // already been performed (headers declared via 1.1 are 
            // implicitly required

            Set<ApplicationData> data = ad.get(elementType);

            // extract the data from module and map it to abstract message
            // part. note that a shortcoming of wsdl 1.1 soap binding is
            // that a header part may map to an elementType of which, multiple
            // instances may exist within message. we'll map first occurrence
            // and leave the rest in the application data bucket

            Iterator<ApplicationData> itr = data.iterator();
            Content temp = (Content)itr.next();
            itr.remove(); // remove it from application data module

            // note that we cannot use the importContent method on
            // ContentFactory. XMLStreamReader is shared by sibling
            // application data elements, i.e. parent is module
            // wrapper element, and parser can only be accessed once,
            // importing the first header would work, importing the
            // second will fail, because parser was accessed

            Node imported = ((Document)message).adoptNode(temp);
            message.setContent(soapHeaderPart, (Content)imported);

        }

        log.trace("readWSDL11DocumentStyleMessage end");
        return message;

    }

    /*
     * Provided for backwards compatability with wsdl 1.1. Binds a runtime abstract message
     * instance using binding rules imported from a wsdl 1.1 definition.
     */
    private static Message readWSDL11RPCStyleMessage(SOAPVersion sv, InterfaceAction ia, SOAPMessage cm, InterfaceMessage am, Message message, MessageModules messageModules, Element content) throws SOAPFault {
        throw new SOAPSenderFault(sv, "RPC-Lit style messages are not currently supported.");
    }

}
