/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util;

import javax.xml.namespace.QName;

import org.apache.axiom.om.OMAttribute;
import org.apache.axiom.om.OMConstants;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMFactory;
import org.apache.axiom.om.OMNamespace;
import org.apache.axiom.soap.SOAP12Version;
import org.apache.axiom.soap.SOAPConstants;
import org.apache.axiom.soap.SOAPEnvelope;
import org.apache.axiom.soap.SOAPFactory;
import org.apache.axiom.soap.SOAPHeader;
import org.bluestemsoftware.specification.eoa.component.application.rt.ApplicationRT;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointActionReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceMessageReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceMessageReference.ContentModel;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageContext;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.SystemFault;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext.MessageModules;
import org.bluestemsoftware.specification.eoa.component.message.InterfaceMessage;
import org.bluestemsoftware.specification.eoa.component.message.MessagePart;
import org.bluestemsoftware.specification.eoa.component.message.rt.Content;
import org.bluestemsoftware.specification.eoa.component.message.rt.ContentFactory;
import org.bluestemsoftware.specification.eoa.component.message.rt.Message;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPDocumentStyleMessage;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPHeaderBlock;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessage;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPMessageStyle;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPVersion;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.appdata.ApplicationData;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.appdata.WSAD.WSAD10;
import org.bluestemsoftware.specification.eoa.system.System.Log;
import org.bluestemsoftware.specification.eoa.system.SystemContext;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

public class SOAPWriter {

    private static Log log = SystemContext.getContext().getSystem().getLog(SOAPWriter.class);
    
    public static void writeMessage(EndpointActionReference ear, InterfaceMessage am, SOAPMessage cm, SOAPVersion soapVersion, MessageContext messageContext, SOAPEnvelope soapEnvelope) throws SystemFault {
        log.trace("writeMessage begin");
        
        MessageModules messageModules = messageContext.getMessageModules();
        Message message = messageContext.getMessage();

        InterfaceMessageReference imr = (InterfaceMessageReference)ear.getReferencedComponent()
                .getInterfaceAction();

        // if the abstract message is null it's a wsdl 20 style message and
        // content model is either #none or #any (it will not be null)
        
        if (am == null) {
            ContentModel contentModel = imr.getContentModel();
            if (contentModel == ContentModel.NONE) {
                if (message != null) {
                    ApplicationRT app = ear.getRootComponent().getApplication();
                    throw new SystemFault(app, "ContentModel #none, but runtime message found.");
                } else {
                    return;
                }
            } else if (contentModel == ContentModel.ANY) {
                if (message.getContent() == null) {
                    ApplicationRT app = ear.getRootComponent().getApplication();
                    throw new SystemFault(app, "ContentModel #any, but payload has no content.");
                }
            }
        }
        
        // if concrete message is null, either the binding is re-usable or
        // the operation is unbound

        if (cm == null) {
            implementDefaultBindingRules(ear, am, messageModules, message, soapEnvelope);
            return;
        }

        // a concrete message definition exists. bind the abstract message
        // according to binding specific rules
        
        if (cm.getMessageStyle() == SOAPMessageStyle.DOCUMENT_STYLE) {
            writeDocumentStyleMessage(ear, imr, cm, am, messageModules, message, soapEnvelope);
        } else {
            writeWSDL11RPCStyleMessage(ear, cm, am, messageModules, message, soapEnvelope);
        }
        
        // note that if any application data exists, application data handler
        // will write to envelope as soap header blocks

        log.trace("writeMessage end");
    }

    /*
     * Implements WSDL 2.0 default binding rules. Note that logic is only executed if binding
     * definition was imported via WSDL 2.0, i.e. when concrete message definition is null (a
     * concrete message def is absolutely required for bindings described via WSDL 1.1.)
     */
    private static void implementDefaultBindingRules(EndpointActionReference ear, InterfaceMessage am, MessageModules messageModules, Message message, SOAPEnvelope soapEnvelope) throws SystemFault {
        log.trace("implementDefaultBindingRules begin");
        
        ApplicationRT app = ear.getRootComponent().getApplication();

        // it's a normal message. wsdl 20 allows for different content
        // models which we must enforce. note that we could let receiver
        // handle validation, but this is more efficient, i.e. avoids
        // unnecessary processing

        InterfaceAction ia = ear.getReferencedComponent().getInterfaceAction();
        ContentModel contentModel = ((InterfaceMessageReference)ia).getContentModel();
        Content content = message.getContent();
        if (contentModel == ContentModel.NONE) {
            if (content != null) {
                throw new SystemFault(app, "ContentModel #none, but payload has content.");
            }
        } else if (contentModel == ContentModel.ANY) {
            if (content == null) {
                throw new SystemFault(app, "ContentModel #any, but payload has no content.");
            } else {
                setBodyContent(app, soapEnvelope, content);
            }
        } else {
            if (content == null) {
                throw new SystemFault(app, "Invalid payload. Expected type "
                        + am.getName()
                        + ", but payload is empty.");
            }
            QName payloadType = new QName(content.getNamespaceURI(), content.getLocalName());
            if (!payloadType.equals(am.getName())) {
                throw new SystemFault(app, "Invalid payload. Expected type "
                        + am.getName()
                        + ", but found type "
                        + payloadType
                        + ".");
            } else {
                setBodyContent(app, soapEnvelope, content);
            }
        }

        // we're done implementing default binding rules. note that
        // there are no declared headers, i.e. no concrete message
        // definition -> no declared headers

        log.trace("implementDefaultBindingRules end");
    }

    private static void writeDocumentStyleMessage(EndpointActionReference ear, InterfaceMessageReference imr, SOAPMessage cm, InterfaceMessage am, MessageModules messageModules, Message message, SOAPEnvelope soapEnvelope) throws SystemFault {
        log.trace("writeDocumentStyleMessage begin");
        
        ApplicationRT app = ear.getRootComponent().getApplication();

        // if content model is defined, it's a wsdl 2.0 style message, i.e.
        // a single abstract part typed as an element which is msg payload.
        // if content model is null, it's a wsdl 1.1 document style msg

        ContentModel contentModel = imr.getContentModel();
        if (contentModel != null) {
            if (contentModel != ContentModel.ELEMENT) {
                throw new SystemFault(app, "Expected ContentModel #element. Found " + contentModel);
            }
            Content content = message.getContent();
            QName payloadType = new QName(content.getNamespaceURI(), content.getLocalName());
            if (!payloadType.equals(am.getName())) {
                throw new SystemFault(app, "Invalid payload. Expected type "
                        + am.getName()
                        + ", but found type "
                        + payloadType
                        + ".");
            } else {
                setBodyContent(app, soapEnvelope, content);
            }
        } else {
            writeWSDL11DocumentStyleMessage(ear, imr, cm, am, messageModules, message, soapEnvelope);
        }

        // note that if any application data exists, application data handler
        // will write to envelope as soap header blocks

        log.trace("writeDocumentStyleMessage end");
    }

    /*
     * writes a WSDL 1.1 document style message, i.e. message payload is an element. Note that
     * unlike a WSDL 2.0 style message, a WSDL 1.1 document style message MAY have multiple
     * parts, i.e. parts that are mapped soap header block elements.
     */
    private static void writeWSDL11DocumentStyleMessage(EndpointActionReference ear, InterfaceMessageReference imr, SOAPMessage cm, InterfaceMessage am, MessageModules messageModules, Message message, SOAPEnvelope soapEnvelope) throws SystemFault {
        log.trace("writeWSDL11DocumentStyleMessage begin");
        
        ApplicationRT app = ear.getRootComponent().getApplication();

        // wsdl 2.0 uses content model to declare an empty payload. content model
        // is undefined for wsdl 1.1, however, so check for a null soap body part

        String soapBodyPart = ((SOAPDocumentStyleMessage)cm).getBodyPart();

        if (soapBodyPart != null) {

            Content content = message.getContent(soapBodyPart);

            // retrieve abstract message part mapped to soap body and verify that its
            // type is as expected

            MessagePart abstractPart = am.getPart(soapBodyPart);
            QName elementType = abstractPart.getSchemaComponentQName();

            if (content == null) {
                throw new SystemFault(app, "Invalid payload. Expected type "
                        + elementType
                        + ", but message part '"
                        + soapBodyPart
                        + "' is empty.");
            }

            QName payloadType = new QName(content.getNamespaceURI(), content.getLocalName());

            if (!payloadType.equals(elementType)) {
                throw new SystemFault(app, "Invalid payload. Expected type "
                        + elementType
                        + ", but message part '"
                        + soapBodyPart
                        + "' typed as "
                        + payloadType
                        + ".");
            } else {
                setBodyContent(app, soapEnvelope, content);
            }

        } else {

            // if there's an extra part in the message which is supposed
            // to be payload, i.e. it doesn't get processed below, we'll
            // silently ignore it

        }

        // now we process declared soap headers, if any, each of which maps
        // to a part declared on abstract message

        for (SOAPHeaderBlock hb : cm.getHeaders()) {

            // retrieve header block type. note that headers are always typed
            // as an element

            String soapHeaderPart = hb.getPartName();
            MessagePart abstractPart = am.getPart(soapHeaderPart);
            QName headerType = abstractPart.getSchemaComponentQName();

            Content content = message.getContent(soapHeaderPart);
            if (content == null) {
                throw new SystemFault(app, "Expected content for declared header "
                        + headerType
                        + ", but message part '"
                        + soapHeaderPart
                        + "' is empty.");
            }

            QName contentType = new QName(content.getNamespaceURI(), content.getLocalName());

            if (!headerType.equals(contentType)) {
                throw new SystemFault(app, "Invalid header part. Expected type "
                        + headerType
                        + ", but message part '"
                        + soapHeaderPart
                        + "' typed as "
                        + contentType
                        + ".");
            }

            // if user streamed header content, we attempt to import content,
            // i.e. by using underlying parser. if reader is not available
            // i.e. if user built-up content rather than streaming it, or
            // if reader was shared with sibling nodes, i.e. content was
            // sourced from underlying message definition, we will catch
            // exception and use dom import method instead

            Document doc = ((Element)soapEnvelope).getOwnerDocument();
            OMElement soapHeaderBlock = null;
            try {
                soapHeaderBlock = (OMElement)((ContentFactory)doc).importContent(content);
            } catch (Exception ex) {
                soapHeaderBlock = (OMElement)((ContentFactory)doc).importNode(content, true);
            }
            
            // rather than using value of must understand on header definition
            // to set runtime attribute, we allow user to control this by
            // passing header part as an instanceof application data. this is
            // also the only way the role attribute can be set on a header part
            
            if (content instanceof ApplicationData) {
                if (((ApplicationData)content).mustUnderstand()) {
                    setMustUnderstand(app, soapEnvelope.getVersion(), soapHeaderBlock);
                }
                String role = ((ApplicationData)content).getRole();
                if (role != null) {
                    setRole(app, soapEnvelope.getVersion(), soapHeaderBlock, role);
                }
                ((Element)soapHeaderBlock).removeAttributeNS(OMConstants.XMLNS_NS_URI, WSAD10.NAMESPACE_PREFIX);
            }

            SOAPHeader soapHeader = soapEnvelope.getHeader();
            if (soapHeader == null) {
                soapHeader = ((SOAPFactory)soapEnvelope.getOMFactory()).createSOAPHeader(soapEnvelope);
            }

            soapHeader.addChild((OMElement)fixUpNamespace((Element)soapHeaderBlock));

        }

        // note that if any application data exists, application data handler
        // will write to envelope as soap header blocks

        log.trace("writeWSDL11DocumentStyleMessage end");
    }

    /*
     * Provided for backwards compatability with wsdl 1.1. Binds a runtime abstract message
     * instance using binding rules imported from a wsdl 1.1 definition.
     */
    private static void writeWSDL11RPCStyleMessage(EndpointActionReference ear, SOAPMessage cm, InterfaceMessage am, MessageModules messageModules, Message message, SOAPEnvelope soapEnvelope) throws SystemFault {
        throw new SystemFault(ear.getRootComponent().getApplication(),
                "RPC-Lit style messages are not currently supported.");
    }

    private static void setBodyContent(ApplicationRT app, SOAPEnvelope soapEnvelope, Content content) throws SystemFault {
        ContentFactory cf = (ContentFactory)((Element)soapEnvelope).getOwnerDocument();
        Element imported = null;
        try {
            imported = cf.importContent(content);
        } catch (Exception ex) {
            imported = (Element)cf.importNode(content, true);
        }
        soapEnvelope.getBody().addChild((OMElement)fixUpNamespace(imported));
    }

    private static void setMustUnderstand(ApplicationRT app, org.apache.axiom.soap.SOAPVersion soapVersion, OMElement header) throws SystemFault {
        header.removeAttribute(header.getAttribute(WSAD10.Infoset.MUST_UNDERSTAND_ATT));
        if (soapVersion instanceof SOAP12Version) {
            setAttribute((OMElement)header, Constants.SOAP_12_MUST_UNDERSTAND_ATTRIB, "true");
        } else {
            setAttribute((OMElement)header, Constants.SOAP_11_MUST_UNDERSTAND_ATTRIB, "1");
        }
    }
    
    private static void setRole(ApplicationRT app, org.apache.axiom.soap.SOAPVersion soapVersion, OMElement header, String role) throws SystemFault {
        header.removeAttribute(header.getAttribute(WSAD10.Infoset.ROLE_ATT));
        if (soapVersion instanceof SOAP12Version) {
            setAttribute((OMElement)header, Constants.SOAP_ROLE_ATTRIB, role);
        } else {
            setAttribute((OMElement)header, Constants.SOAP_ACTOR_ATTRIB, role);
        }
    }

    private static void setAttribute(OMElement parent, QName name, String value) {
        OMAttribute omAttribute = parent.getAttribute(name);
        if (omAttribute != null) {
            omAttribute.setAttributeValue(value);
        } else {
            OMFactory factory = parent.getOMFactory();
            OMNamespace omNamespace = parent.findNamespace(name.getNamespaceURI(), null);
            if (omNamespace == null) {
                String prefix = SOAPConstants.SOAP_DEFAULT_NAMESPACE_PREFIX;
                omNamespace = factory.createOMNamespace(name.getNamespaceURI(), prefix);
            }
            omAttribute = factory.createOMAttribute(name.getLocalPart(), omNamespace, value);
            parent.addAttribute(omAttribute);
        }
    }

    private static final String XMLNS_NS = "http://www.w3.org/2000/xmlns/";

    private static Element fixUpNamespace(Element element) {
        if (element.getPrefix() == null || element.getPrefix().equals("")) {
            element.setAttributeNS(XMLNS_NS, "xmlns", element.getNamespaceURI());
        } else {
            element.setAttributeNS(XMLNS_NS, "xmlns:" + element.getPrefix(), element.getNamespaceURI());
        }
        return element;
    }

}
