/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.wsa;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMNamespace;
import org.apache.axiom.om.util.ElementHelper;
import org.apache.axiom.soap.SOAPEnvelope;
import org.apache.axiom.soap.SOAPFactory;
import org.apache.axiom.soap.SOAPHeader;
import org.apache.axiom.soap.SOAPHeaderBlock;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.HeadersUtil;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext.MessageModules;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPVersion;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPFault;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.WSA;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.WSAMessageModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.WSA.WSA10;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.soap.fault.InvalidEPRFault;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.soap.fault.MissingAddressInEPRFault;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.addressing.soap.fault.WSAddressingFault;
import org.bluestemsoftware.specification.wsa.EndpointReference;
import org.w3c.dom.Element;

public class WSA10Handler extends WSAHandler {

    private static WSAHandler singleton = new WSA10Handler();

    public static WSAHandler getInstance() {
        return singleton;
    }

    @Override
    public WSAMessageModule handleMyServiceRequest(SOAPEnvelope message) throws SOAPFault {
        return readMessage(message, false);
    }

    @Override
    public WSAMessageModule handleMyServiceResponse(MessageModules requestModules, MessageModules responseModules, SOAPEnvelope message) throws SOAPFault {

        WSAMessageModule wsammIn = requestModules.getModule(WSAMessageModule.class);
        WSAMessageModule wsammOut = responseModules.getModule(WSAMessageModule.class);

        EndpointReference endpointReference = null;
        if (message.getBody().getFirstElement() instanceof org.apache.axiom.soap.SOAPFault) {
            if (wsammIn.getFaultTo() != null) {
                endpointReference = wsammIn.getFaultTo();
            } else {
                endpointReference = wsammIn.getReplyTo();
            }
        } else {
            endpointReference = wsammIn.getReplyTo();
        }

        String to = endpointReference.getAddress();

        // if to address starts with 'jms', replace this with anonymous address,
        // i.e. binding uses a jms replyto to simulate synchronous behavior

        if (to.startsWith("jms")) {
            to = WSA10.ANONYMOUS_URI;
        }

        wsammOut.setTo(to);
        Set<Element> refParms = endpointReference.getReferenceParameters();

        SOAPFactory factory = (SOAPFactory)message.getOMFactory();
        SOAPHeader header = message.getHeader();

        OMNamespace ns = factory.createOMNamespace(WSA10.NAMESPACE_URI, WSA.NAMESPACE_PREFIX);
        
        SOAPHeaderBlock toHeaderBlock = header.addHeaderBlock(WSA.TO, ns);
        toHeaderBlock.setText(to);

        SOAPHeaderBlock actionHeaderBlock = header.addHeaderBlock(WSA.ACTION, ns);
        actionHeaderBlock.setText(wsammOut.getAction());
        
        // by convention message id is omitted from headers by some providers,
        // e.g. axis. we, however, include it for two reasons. first, there's
        // no need to generate a unique string on both ends, second, if partner
        // is on same machine, i.e. message transported in vm, the message id is
        // required to look-up message modules within receive partner service
        // response method. btw, wsa 1.0 core spec includes message id header in
        // 'formulating a reply message' example
        
        SOAPHeaderBlock messageIDHeaderBlock = header.addHeaderBlock(WSA.MESSAGE_ID, ns);
        messageIDHeaderBlock.setText(wsammOut.getMessageID());

        SOAPHeaderBlock relatesToHeaderBlock = header.addHeaderBlock(WSA.RELATES_TO, ns);
        String relatesTo = wsammIn.getMessageID();
        wsammOut.setRelatesTo(relatesTo);
        relatesToHeaderBlock.setText(wsammIn.getMessageID());

        if (refParms.size() > 0) {
            SOAPHeaderBlock rpHeaderBlock = header.addHeaderBlock(WSA.REF_PARMS, ns);
            for (Element rp : refParms) {
                rpHeaderBlock.addChild(ElementHelper.importOMElement((OMElement)rp, factory));
            }
            wsammOut.setReferenceParameters(refParms);
        }

        return wsammOut;
    }

    @Override
    public void handleWSAFaultResponse(String to, String relatesTo, String action, SOAPEnvelope message) {

        to = to == null ? WSA10.ANONYMOUS_URI : to;
        
        SOAPFactory factory = (SOAPFactory)message.getOMFactory();
        SOAPHeader header = message.getHeader();

        OMNamespace ns = factory.createOMNamespace(WSA10.NAMESPACE_URI, WSA.NAMESPACE_PREFIX);
        
        SOAPHeaderBlock toHeaderBlock = header.addHeaderBlock(WSA.TO, ns);
        toHeaderBlock.setText(to);

        SOAPHeaderBlock actionHeaderBlock = header.addHeaderBlock(WSA.ACTION, ns);
        actionHeaderBlock.setText(action);

        SOAPHeaderBlock relatesToHeaderBlock = header.addHeaderBlock(WSA.RELATES_TO, ns);
        relatesToHeaderBlock.setText(relatesTo);
        
    }

    
    @Override
    public WSAMessageModule handlePartnerServiceResponse(SOAPEnvelope message) throws SOAPFault {
        return readMessage(message, true);
    }

    @Override
    public void handlePartnerServiceRequest(MessageModules requestModules, SOAPEnvelope message) throws SOAPFault {

        // retrieve values populated by caller which were derived using policy
        // retrieved from supported versions of ws-a spec defined on binding
        // module defined on partner action. 

        WSAMessageModule messageModule = requestModules.getModule(WSAMessageModule.class);
        String to = messageModule.getTo();
        String action = messageModule.getAction();
        String messageID = messageModule.getMessageID();
        EndpointReference replyTo = messageModule.getReplyTo();
        EndpointReference faultTo = messageModule.getFaultTo();

        SOAPFactory factory = (SOAPFactory)message.getOMFactory();
        SOAPHeader header = message.getHeader();

        OMNamespace ns = factory.createOMNamespace(WSA10.NAMESPACE_URI, WSA.NAMESPACE_PREFIX);
        
        SOAPHeaderBlock toHeaderBlock = header.addHeaderBlock(WSA.TO, ns);
        toHeaderBlock.setText(to);

        SOAPHeaderBlock actionHeaderBlock = header.addHeaderBlock(WSA.ACTION, ns);
        actionHeaderBlock.setText(action);

        SOAPHeaderBlock messageIDHeaderBlock = header.addHeaderBlock(WSA.MESSAGE_ID, ns);
        messageIDHeaderBlock.setText(messageID);

        SOAPHeaderBlock replyToHeaderBlock = header.addHeaderBlock(WSA.REPLY_TO, ns);
        OMElement addressElement = factory.createOMElement(WSA10.Infoset.ADDRESS_ELEM);
        replyToHeaderBlock.addChild(addressElement);
        addressElement.setText(replyTo.getAddress());

        if (faultTo != null) {
            SOAPHeaderBlock faultToHeaderBlock = header.addHeaderBlock(WSA.FAULT_TO, ns);
            addressElement = factory.createOMElement(WSA10.Infoset.ADDRESS_ELEM);
            faultToHeaderBlock.addChild(addressElement);
            addressElement.setText(faultTo.getAddress());
        }

        // add reference parameters associated with destination epr, i.e. epr
        // retrieved by caller which correllates with To url

        Set<Element> refParms = messageModule.getReferenceParameters();
        if (refParms.size() > 0) {
            SOAPHeaderBlock rpHeaderBlock = header.addHeaderBlock(WSA.REF_PARMS, ns);
            for (Element rp : refParms) {
                rpHeaderBlock.addChild(ElementHelper.importOMElement((OMElement)rp, factory));
            }
        }

    }

    private WSAMessageModule readMessage(SOAPEnvelope message, boolean isResponse) throws WSAddressingFault {

        String envelopeNamespace = message.getVersion().getEnvelopeURI();
        SOAPVersion sv = SOAPVersion.valueOfEnvelopeNamespace(envelopeNamespace);

        // retrieve all headers targeted to ultimate destination, i.e.
        // which is assumed when we pass a null argument

        Iterator<OMElement> itr = HeadersUtil.getHeadersForUltimateDestination(sv, message.getHeader(), null);

        String action = null;
        String to = null;
        String messageID = null;
        EndpointReference replyTo = null;
        EndpointReference faultTo = null;
        EndpointReference from = null;
        String relatesTo = null;
        Set<Element> referenceParameters = null;

        // attempt to parse as many valid properties as possible before throwing
        // any ws-addressing faults, i.e. to aid with fault response processing.
        // note that fault may be overwritten if more than one problem exists,
        // in which case we end up throwing the last problem encountered

        WSAddressingFaultImpl wsaf = null;

        while (itr.hasNext()) {

            OMElement hb = itr.next();

            if (hb.getNamespace() == null) {
                continue;
            }

            if (!hb.getNamespace().getNamespaceURI().equals(WSA10.NAMESPACE_URI)) {
                continue;
            }

            if (hb.getLocalName().equals(WSA.ACTION)) {
                if (action != null) {
                    wsaf = new InvalidCardinalityFaultImpl(sv, WSA10.NAMESPACE_URI, WSA.ACTION);
                } else {
                    action = hb.getText();
                    hb.detach();
                    continue;
                }
            }

            if (hb.getLocalName().equals(WSA.REPLY_TO)) {
                if (replyTo != null) {
                    wsaf = new InvalidCardinalityFaultImpl(sv, WSA10.NAMESPACE_URI, WSA.REPLY_TO);
                } else {
                    try {
                        replyTo = parseEndpointReference(sv, hb, WSA.REPLY_TO);
                    } catch (InvalidEPRFault invalidEPRFault) {
                        wsaf = (WSAddressingFaultImpl)invalidEPRFault;
                    }
                    hb.detach();
                    continue;
                }
            }

            if (hb.getLocalName().equals(WSA.MESSAGE_ID)) {
                if (messageID != null) {
                    wsaf = new InvalidCardinalityFaultImpl(sv, WSA10.NAMESPACE_URI, WSA.MESSAGE_ID);
                } else {
                    messageID = hb.getText();
                    hb.detach();
                    continue;
                }
            }

            if (hb.getLocalName().equals(WSA.RELATES_TO)) {
                if (relatesTo != null) {
                    wsaf = new InvalidCardinalityFaultImpl(sv, WSA10.NAMESPACE_URI, WSA.RELATES_TO);
                } else {
                    String mid = hb.getText();
                    String rt = hb.getAttributeValue(WSA10.Infoset.RELATIONSHIP_TYPE_ATT);
                    if (rt != null && !rt.equals(WSA10.DEFAULT_RELATIONSHIP_TYPE)) {
                        throw new InvalidHeaderFaultImpl(sv, WSA10.NAMESPACE_URI, WSA.RELATIONSHIP_TYPE);
                    }
                    relatesTo = mid;
                    hb.detach();
                    continue;
                }
            }

            if (hb.getLocalName().equals(WSA.FAULT_TO)) {
                if (faultTo != null) {
                    wsaf = new InvalidCardinalityFaultImpl(sv, WSA10.NAMESPACE_URI, WSA.FAULT_TO);
                } else {
                    try {
                        faultTo = parseEndpointReference(sv, hb, WSA.FAULT_TO);
                    } catch (InvalidEPRFault invalidEPRFault) {
                        wsaf = (WSAddressingFaultImpl)invalidEPRFault;
                    }
                    hb.detach();
                    continue;
                }
            }

            if (hb.getLocalName().equals(WSA.TO)) {
                if (to != null) {
                    wsaf = new InvalidCardinalityFaultImpl(sv, WSA10.NAMESPACE_URI, WSA.TO);
                } else {
                    to = hb.getText();
                    hb.detach();
                    continue;
                }
            }

            if (hb.getLocalName().equals(WSA.FROM)) {
                if (from != null) {
                    wsaf = new InvalidCardinalityFaultImpl(sv, WSA10.NAMESPACE_URI, WSA.FROM);
                } else {
                    try {
                        from = parseEndpointReference(sv, hb, WSA.FROM);
                    } catch (InvalidEPRFault invalidEPRFault) {
                        wsaf = (WSAddressingFaultImpl)invalidEPRFault;
                    }
                    hb.detach();
                    continue;
                }
            }

            if (hb.getLocalName().equals(WSA.REF_PARMS)) {
                if (referenceParameters != null) {
                    wsaf = new InvalidCardinalityFaultImpl(sv, WSA10.NAMESPACE_URI, WSA.REF_PARMS);
                } else {
                    referenceParameters = new HashSet<Element>();
                    Iterator<?> itr2 = hb.getChildElements();
                    while (itr2.hasNext()) {
                        OMElement rp = (OMElement)itr2.next();
                        referenceParameters.add((Element)rp.detach());
                    }
                    hb.detach();
                    continue;
                }
            }

        }

        if (action == null) {
            wsaf = new HeaderRequiredFaultImpl(sv, WSA10.NAMESPACE_URI, WSA.ACTION);
        }

        if (isResponse) {

            if (to == null) {
                to = WSA10.ANONYMOUS_URI;
            }

            if (relatesTo == null || relatesTo.equals("")) {
                wsaf = new InvalidHeaderFaultImpl(sv, WSA10.NAMESPACE_URI, WSA.RELATES_TO);
            }

            // if message id was not provided, it will be generated
            // within module constructor

        } else {

            if (replyTo == null) {
                replyTo = WSA10.DEFAULT_REPLY_TO;
            }

            if (messageID == null || messageID.equals("")) {
                messageID = WSA10.UNSPECIFIED_MESSAGE;
                wsaf = new InvalidHeaderFaultImpl(sv, WSA10.NAMESPACE_URI, WSA.MESSAGE_ID);
            }

            if (to == null) {
                wsaf = new HeaderRequiredFaultImpl(sv, WSA10.NAMESPACE_URI, WSA.TO);
            }

        }

        WSAMessageModule wsaMessageModule = null;
        if (messageID == null) {
            wsaMessageModule = new WSAMessageModule(action);
        } else {
            wsaMessageModule = new WSAMessageModule(action, messageID);
        }

        wsaMessageModule.setFaultTo(faultTo);
        wsaMessageModule.setFrom(from);
        wsaMessageModule.setNamespace(WSA10.NAMESPACE_URI);
        wsaMessageModule.setReferenceParameters(referenceParameters);
        wsaMessageModule.setRelatesTo(relatesTo);
        wsaMessageModule.setReplyTo(replyTo);
        wsaMessageModule.setTo(to);

        if (wsaf != null) {
            wsaf.setOffendingMessageModule(wsaMessageModule);
            throw (WSAddressingFault)wsaf;
        }

        return wsaMessageModule;

    }

    private static EndpointReference parseEndpointReference(SOAPVersion sv, OMElement hb, String headerName) throws MissingAddressInEPRFault {

        // note that from, replyTo and faultTo are all typed as complex
        // schema type wsa:EndpointReferenceType, i.e. the parent element
        // is a wildcard. only the child elements are defined by grammar

        OMElement addElement = hb.getFirstChildWithName(WSA10.Infoset.ADDRESS_ELEM);
        if (addElement == null) {
            throw new MissingAddressInEPRFaultImpl(sv, WSA10.NAMESPACE_URI, headerName);
        }
        String address = addElement.getText();
        if (address == null || address.equals("")) {
            throw new MissingAddressInEPRFaultImpl(sv, WSA10.NAMESPACE_URI, headerName);
        }
        try {
            new URI(address);
        } catch (URISyntaxException ue) {
            throw new InvalidEPRFault(sv, WSA10.NAMESPACE_URI, headerName);
        }
        Set<Element> refParams = null;
        OMElement refParmsElement = hb.getFirstChildWithName(WSA10.Infoset.REF_PARMS_ELEM);
        if (refParmsElement != null) {
            Iterator<?> itr2 = refParmsElement.getChildElements();
            while (itr2.hasNext()) {
                OMElement rp = (OMElement)itr2.next();
                refParams.add((Element)rp.detach());
            }
            return new EndpointReference(address, refParams);
        } else {
            return new EndpointReference(address);
        }

    }

}
