/**
 * Copyright 2008 Bluestem Software LLC.  All Rights Reserved.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

package org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.wsad;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.apache.axiom.om.OMAttribute;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMFactory;
import org.apache.axiom.om.OMNamespace;
import org.apache.axiom.soap.SOAPEnvelope;
import org.apache.axiom.soap.SOAPFactory;
import org.apache.axiom.soap.SOAPHeader;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.Constants;
import org.bluestemsoftware.open.eoa.ext.binding.soap.dfault.util.HeadersUtil;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext.MessageModules;
import org.bluestemsoftware.specification.eoa.component.message.rt.Content;
import org.bluestemsoftware.specification.eoa.component.message.rt.ContentFactory;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.SOAPVersion;
import org.bluestemsoftware.specification.eoa.ext.binding.soap.rt.SOAPFault;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.appdata.ApplicationData;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.appdata.ApplicationDataModule;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.appdata.WSAD.WSAD10;

public class ApplicationDataHandler {

    public static Map<QName, Set<ApplicationData>> handleMyServiceRequest(String role, Set<QName> understoodHeaders, SOAPEnvelope message) throws SOAPFault {
        return readMessage(role, understoodHeaders, message);
    }

    public static void handleMyServiceResponse(MessageModules responseModules, SOAPEnvelope message) throws SOAPFault {

        ApplicationDataModule applicationDataModule = responseModules.getModule(ApplicationDataModule.class);
        Map<QName, Set<ApplicationData>> applicationData = applicationDataModule.getApplicationData();

        if (applicationData.size() > 0) {
            writeMessage(message, applicationData);
        }

    }

    public static Map<QName, Set<ApplicationData>> handlePartnerServiceResponse(String role, Set<QName> understoodHeaders, SOAPEnvelope message) throws SOAPFault {
        return readMessage(role, understoodHeaders, message);
    }

    public static void handlePartnerServiceRequest(MessageModules requestModules, SOAPEnvelope message) throws SOAPFault {

        ApplicationDataModule applicationDataModule = requestModules.getModule(ApplicationDataModule.class);
        Map<QName, Set<ApplicationData>> applicationData = applicationDataModule.getApplicationData();

        if (applicationData.size() > 0) {
            writeMessage(message, applicationData);
        }

    }

    private static Map<QName, Set<ApplicationData>> readMessage(final String role, Set<QName> understoodHeaders, SOAPEnvelope message) throws SOAPFault {

        String envelopeNamespace = message.getVersion().getEnvelopeURI();
        SOAPVersion sv = SOAPVersion.valueOfEnvelopeNamespace(envelopeNamespace);

        Iterator<OMElement> itr = HeadersUtil.getHeadersForUltimateDestination(sv, message.getHeader(), role);

        if (!itr.hasNext()) {
            return null;
        }

        String soapNamespace = message.getNamespace().getNamespaceURI();

        Map<QName, Set<ApplicationData>> applicationData = null;

        // if header is flagged as must understand, we must not extract
        // if not understood by service, i.e. as configured via policy.
        // if not flagged as must understand, or header is understood,
        // detach the header block and remove all protocol specific
        // attributes as required by eoa spec, i.e. the message module
        // is abstract

        while (itr.hasNext()) {
            OMElement hb = itr.next();
            String value = null;
            if (sv == SOAPVersion.SOAP11) {
                value = hb.getAttributeValue(Constants.SOAP_11_MUST_UNDERSTAND_ATTRIB);
            } else {
                value = hb.getAttributeValue(Constants.SOAP_12_MUST_UNDERSTAND_ATTRIB);
            }
            boolean mustUnderstand = false;
            if (value != null) {
                value = value.equals("1") ? "true" : value.equals("0") ? "false" : value;
                mustUnderstand = Boolean.parseBoolean(value);
            }
            if (mustUnderstand) {
                QName headerName = hb.getQName();
                if (!understoodHeaders.contains(headerName)) {
                    continue;
                }
            }
            hb.detach();
            Iterator<?> attribs = hb.getAllAttributes();
            while (attribs.hasNext()) {
                OMAttribute attrib = (OMAttribute)attribs.next();
                String namespace = attrib.getQName().getNamespaceURI();
                if (namespace != null && namespace.equals(soapNamespace)) {
                    hb.removeAttribute(attrib);
                }
            }
            if (applicationData == null) {
                applicationData = new HashMap<QName, Set<ApplicationData>>();
            }
            Set<ApplicationData> set = applicationData.get(hb.getQName());
            if (set == null) {
                set = new HashSet<ApplicationData>();
                applicationData.put(hb.getQName(), set);
            }
            set.add((ApplicationData)hb);
        }

        return applicationData;

    }

    private static void writeMessage(SOAPEnvelope message, Map<QName, Set<ApplicationData>> applicationData) {

        SOAPHeader soapHeader = message.getHeader();
        if (soapHeader == null) {
            soapHeader = ((SOAPFactory)message.getOMFactory()).createSOAPHeader(message);
        }

        ContentFactory contentFactory = ((Content)soapHeader).getFactory();

        // iterate over application data and write a header block for each. set
        // must understand and role/actor attrib if role not null then remove
        // wsad attribs for role and must understand copied from ad element

        for (Set<ApplicationData> set : applicationData.values()) {
            for (ApplicationData element : set) {
                OMElement hb = (OMElement)contentFactory.importNode(element, true);
                String role = element.getRole();
                if (role != null) {
                    OMAttribute att = hb.getAttribute(WSAD10.Infoset.ROLE_ATT);
                    hb.removeAttribute(att);
                    setRole(message.getVersion(), hb, role);
                }
                if (element.mustUnderstand()) {
                    OMAttribute att = hb.getAttribute(WSAD10.Infoset.MUST_UNDERSTAND_ATT);
                    hb.removeAttribute(att);
                    setMustUnderstand(message.getVersion(), hb);
                }
                soapHeader.addChild(hb);
            }
        }

    }

    private static void setMustUnderstand(org.apache.axiom.soap.SOAPVersion soapVersion, OMElement omElement) {
        if (soapVersion instanceof org.apache.axiom.soap.SOAP12Version) {
            setAttribute(omElement, Constants.SOAP_12_MUST_UNDERSTAND_ATTRIB, "1");
        } else {
            setAttribute(omElement, Constants.SOAP_11_MUST_UNDERSTAND_ATTRIB, "1");
        }
    }

    private static void setRole(org.apache.axiom.soap.SOAPVersion soapVersion, OMElement omElement, String role) {
        if (soapVersion instanceof org.apache.axiom.soap.SOAP12Version) {
            setAttribute(omElement, Constants.SOAP_ROLE_ATTRIB, role);
        } else {
            setAttribute(omElement, Constants.SOAP_ACTOR_ATTRIB, role);
        }
    }

    private static void setAttribute(OMElement parent, QName name, String value) {
        OMAttribute omAttribute = parent.getAttribute(name);
        if (omAttribute != null) {
            omAttribute.setAttributeValue(value);
        } else {
            OMFactory factory = parent.getOMFactory();
            OMNamespace omNamespace = parent.findNamespace(name.getNamespaceURI(), null);
            if (omNamespace == null) {
                String prefix = org.apache.axiom.soap.SOAPConstants.SOAP_DEFAULT_NAMESPACE_PREFIX;
                omNamespace = factory.createOMNamespace(name.getNamespaceURI(), prefix);
            }
            omAttribute = factory.createOMAttribute(name.getLocalPart(), omNamespace, value);
            parent.addAttribute(omAttribute);
        }
    }

}
